'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _index = require('../utils/index');

var ParallaxScroller = function () {

    function ParallaxScrollListener() {

        // All parallax elements to be updated
        var elements = [];

        // Tracks current scroll y distance
        var scrollY = 0;

        // ID to increment for elements
        var id = 0;

        // Ticking
        var ticking = false;

        // Scroll direction
        var scrollDown = null;

        // Passive support
        var supportsPassive = (0, _index.testForPassiveScroll)();

        function _removeListeners() {
            window.removeEventListener('scroll', _handleScroll, supportsPassive ? { passive: true } : false);
            window.removeEventListener('resize', _handleResize, false);
        }

        function _addListeners() {
            window.addEventListener('scroll', _handleScroll, supportsPassive ? { passive: true } : false);
            window.addEventListener('resize', _handleResize, false);
        }

        /**
         * Window scroll handler
         * sets the scrollY and then calls updateElementPositions()
         */
        function _handleScroll() {
            // reference to prev scroll y
            var prevScrollY = scrollY;

            // Save current scroll
            scrollY = window.pageYOffset; // Supports IE 9 and up.

            // direction
            scrollDown = scrollY > prevScrollY;

            // Only called if the last animation request has been
            // completed and there are parallax elements to update
            if (!ticking && elements.length > 0) {
                ticking = true;
                window.requestAnimationFrame(_updateElementPositions);
            }
        }

        /**
         * Window resize handler
         * calls update() which update parallax element attributes and positions
         */
        function _handleResize() {
            _updateElementAttributes();
            _updateElementPositions();
        }

        /**
         * Creates a unique ID
         */
        function _createID() {
            ++id;
            return id;
        }

        function _updateElementPositions() {
            elements.forEach(function (element) {
                if (element.props.disabled) return;

                // check if the element is in view then
                var isInView = _isElementInView(element);

                // set styles if it is
                if (isInView) _setParallaxStyles(element);

                // reset ticking so more animations can be called
                ticking = false;
            });
        }

        function _updateElementAttributes() {
            elements.forEach(function (element) {
                if (element.props.disabled) return;

                _setupOffsets(element);

                _cacheAttributes(element);
            });
        }

        function _removeParallaxStyles() {
            elements.forEach(function (element) {
                _resetStyles(element);
            });
        }

        /**
         * Takes a parallax element and caches important values
         * as an attribute object on the element
         *
         * @param {object} element
         */
        function _cacheAttributes(element) {
            var _element$offsets = element.offsets,
                yMin = _element$offsets.yMin,
                yMax = _element$offsets.yMax,
                xMax = _element$offsets.xMax,
                xMin = _element$offsets.xMin;
            var slowerScrollRate = element.props.slowerScrollRate;

            // NOTE: Many of these cause layout and reflow so we're not
            // calculating them on every frame -- instead these values
            // are cached on the element to access later when determining
            // the element's position and offset.

            var el = element._outer;
            var rect = el.getBoundingClientRect();
            var html = document.documentElement;
            var windowHeight = window.innerHeight || html.clientHeight;
            var elHeight = el.offsetHeight;
            var elWidth = el.offsetWidth;
            var scrollY = window.pageYOffset;

            // NOTE: offsetYMax and offsetYMin are percents
            // based of the height of the element. They must be
            // calculated as px to correctly determine whether
            // the element is in the viewport.
            var yPercent = yMax.unit === '%' || yMin.unit === '%';
            var xPercent = xMax.unit === '%' || xMin.unit === '%';

            // X offsets
            var yMinPx = yMin.value;
            var yMaxPx = yMax.value;

            if (yPercent) {
                var h100 = elHeight / 100;
                yMaxPx = yMax.value * h100;
                yMinPx = yMin.value * h100; // negative value
            }

            // Y offsets
            var xMinPx = xMax.value;
            var xMaxPx = xMin.value;

            if (xPercent) {
                var w100 = elWidth / 100;
                xMaxPx = xMax.value * w100;
                xMinPx = xMin.value * w100; // negative value
            }

            // NOTE: must add the current scroll position when the
            // element is checked so that we get its absolute position
            // relative to the document and not the viewport then
            // add the min/max offsets calculated above
            var top = 0;
            var bottom = 0;

            if (slowerScrollRate) {
                top = rect.top + scrollY + yMinPx;
                bottom = rect.bottom + scrollY + yMaxPx;
            } else {
                top = rect.top + scrollY + yMaxPx * -1;
                bottom = rect.bottom + scrollY + yMinPx * -1;
            }

            // Total distance the element will move from when
            // the top enters the view to the bottom leaving
            // accounting for elements height and max/min offsets
            var totalDist = windowHeight + (elHeight + Math.abs(yMinPx) + yMaxPx);

            element.attributes = {
                top: top,
                bottom: bottom,
                elHeight: elHeight,
                elWidth: elWidth,
                yMaxPx: yMaxPx,
                yMinPx: yMinPx,
                xMaxPx: xMaxPx,
                xMinPx: xMinPx,
                totalDist: totalDist,
                windowHeight: windowHeight
            };
        }

        /**
         * Takes a parallax element and parses the offset props
         * to get the value and unit (if any). Sets these offsets
         * on the element
         *
         * @param {object} element
         */
        function _setupOffsets(element) {
            var _element$props = element.props,
                offsetYMin = _element$props.offsetYMin,
                offsetYMax = _element$props.offsetYMax,
                offsetXMax = _element$props.offsetXMax,
                offsetXMin = _element$props.offsetXMin;


            var yMin = (0, _index.parseValueAndUnit)(offsetYMin);
            var yMax = (0, _index.parseValueAndUnit)(offsetYMax);
            var xMin = (0, _index.parseValueAndUnit)(offsetXMax);
            var xMax = (0, _index.parseValueAndUnit)(offsetXMin);

            // @TODO: Move error to component proptypes
            if (xMin.unit !== xMax.unit || yMin.unit !== yMax.unit) {
                throw new Error('Must provide matching units for the min and max offset values of each axis.');
            }

            var xUnit = xMin.unit || '%';
            var yUnit = yMin.unit || '%';

            element.offsets = {
                xUnit: xUnit,
                yUnit: yUnit,
                yMin: yMin,
                yMax: yMax,
                xMin: xMin,
                xMax: xMax
            };
        }

        function _isElementInView(element) {
            var windowHeight = element.attributes.windowHeight;

            var top = element.attributes.top - scrollY;
            var bottom = element.attributes.bottom - scrollY;

            var topInView = top >= 0 && top <= windowHeight;
            var bottomInView = bottom >= 0 && bottom <= windowHeight;
            var covering = top <= 0 && bottom >= windowHeight;

            var isInView = topInView || bottomInView || covering;
            // console.log('top', topInView, 'bottom', bottomInView, 'covering', covering);

            return isInView;
        }

        function _setParallaxStyles(element) {
            var top = element.attributes.top - scrollY;
            var _element$attributes = element.attributes,
                windowHeight = _element$attributes.windowHeight,
                totalDist = _element$attributes.totalDist;

            // Percent the element has moved based on current and total distance to move

            var percentMoved = (top * -1 + windowHeight) / totalDist * 100;

            // Scale percentMoved to min/max percent determined by offset props
            var slowerScrollRate = element.props.slowerScrollRate;

            // Get the parallax X and Y offsets

            var offsets = (0, _index.getParallaxOffsets)(element.offsets, percentMoved, slowerScrollRate);

            // Apply styles
            var el = element._inner;
            el.style.cssText = 'will-change:transform;\n                position:relative;\n                transform:translate3d(' + offsets.x.value + offsets.x.unit + ', ' + offsets.y.value + offsets.y.unit + ', 0)';
        }

        function _resetStyles(element) {
            // Resets any styles that may be left over when
            // resizing from desktop to mobile apply styles
            var el = element._inner;
            el.style.cssText = 'will-change:none;\n                position:relative;\n                transform:translate3d(0, 0, 0)';
        }

        /**
         * --------------------------------------
         * Public methods
         * --------------------------------------
         */

        this.createElement = function (options) {
            var id = _createID();
            var element = _extends({
                id: id
            }, options);

            elements.push(element);
            this.update();

            return element;
        };

        this.removeElement = function (element) {
            // console.log('remove element')
            var index = elements.indexOf(element);
            if (index !== -1) {
                elements.splice(index, 1);
            }
        };

        this.updateElement = function (id, options) {
            // console.log('update element')
            // update props of a given element
            var index = elements.findIndex(function (el) {
                return el.id === id;
            });

            // create new element with options
            var element = Object.assign({}, elements[index], options);

            // replace old
            elements[index] = element;

            this.update();
        };

        /**
         * Updates all parallax element attributes and postitions
         */
        this.update = function () {
            _updateElementAttributes();
            _updateElementPositions();
        };

        /**
         * Removes listeners, resets all styles, nullifies self
         */
        this.reset = function () {
            _removeParallaxStyles();
        };

        /**
         * Removes listeners, resets all styles, nullifies self
         */
        this.destroy = function () {
            _removeListeners();
            _removeParallaxStyles();
            window.ParallaxScrollListener = null;
        };

        // Initialize the Parallax scroll listener
        _addListeners();
    }

    return {
        // Return the singleton instance if one exists
        // or create one if it doesn't
        init: function init() {
            if (!window.ParallaxScrollListener) {
                window.ParallaxScrollListener = new ParallaxScrollListener();
            }
            return window.ParallaxScrollListener;
        }
    };
}();

exports.default = ParallaxScroller;
module.exports = exports['default'];